import datetime
import json
import random

from django import forms
from django.core.exceptions import PermissionDenied
from django.db.models import Sum, Q, F
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext as _, gettext_lazy
import django_filters
from django_filters.constants import EMPTY_VALUES
from wagtail.admin import messages
from wagtail.admin.action_menu import ActionMenuItem
from wagtail.admin.filters import DateRangePickerWidget, WagtailFilterSet
from wagtail.admin.views.reports import ReportView
from wagtail.core.models import Page, PAGE_MODEL_CLASSES, UserPagePermissionsProxy

from .models import AbTest
from .events import EVENT_TYPES


class CreateAbTestForm(forms.ModelForm):
    goal_event = forms.ChoiceField(choices=[])
    hypothesis = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['goal_event'].choices = [
            (slug, goal.name)
            for slug, goal in EVENT_TYPES.items()
        ]

    def save(self, page, variant_revision, user):
        ab_test = super().save(commit=False)
        ab_test.page = page
        ab_test.variant_revision = variant_revision
        ab_test.created_by = user
        ab_test.save()
        return ab_test

    class Meta:
        model = AbTest
        fields = ['name', 'hypothesis', 'goal_event', 'goal_page', 'sample_size']


def add_ab_test_checks(request, page):
    # User must have permission to edit the page
    page_perms = page.permissions_for_user(request.user)
    if not page_perms.can_edit():
        raise PermissionDenied

    # User must have permission to add A/B tests
    if not request.user.has_perm('wagtail_ab_testing.add_abtest'):
        raise PermissionDenied

    # Page must not already be running an A/B test
    if AbTest.objects.get_current_for_page(page=page):
        messages.error(request, _("This page already has a running A/B test"))

        return redirect('wagtailadmin_pages:edit', page.id)

    # Page must be published and have a draft revision
    if not page.live or not page.has_unpublished_changes:
        messages.error(request, _("To run an A/B test on this page, it must be live with draft changes."))

        return redirect('wagtailadmin_pages:edit', page.id)


def add_compare(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    # Run some checks
    response = add_ab_test_checks(request, page)
    if response:
        return response

    latest_revision_as_page = page.get_latest_revision().as_page_object()
    comparison = page.get_edit_handler().get_comparison()
    comparison = [comp(page, latest_revision_as_page) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return render(request, 'wagtail_ab_testing/add_compare.html', {
        'page': page,
        'latest_revision_as_page': latest_revision_as_page,
        'comparison': comparison,
        'differences': any(comp.has_changed() for comp in comparison),
    })


def add_form(request, page_id):
    page = get_object_or_404(Page, id=page_id)

    # Run some checks
    response = add_ab_test_checks(request, page)
    if response:
        return response

    if request.method == 'POST':
        form = CreateAbTestForm(request.POST)

        if form.is_valid():
            ab_test = form.save(page, page.get_latest_revision(), request.user)

            if 'start' in request.POST:
                page_perms = page.permissions_for_user(request.user)
                if page_perms.can_publish():
                    ab_test.start()

                    messages.success(request, _("The A/B test has been created and started."))
                else:
                    messages.error(request, _("The A/B test has been created but you do not have permission to publish so it couldn't be started."))

            else:
                messages.success(request, _("The A/B test has been created."))

            return redirect('wagtailadmin_pages:edit', page.id)
    else:
        form = CreateAbTestForm()

    return render(request, 'wagtail_ab_testing/add_form.html', {
        'page': page,
        'form': form,
        'goal_selector_props': json.dumps({
            'goalTypesByPageType': {
                f'{page_type._meta.app_label}.{page_type._meta.model_name}': [
                    {
                        'slug': slug,
                        'name': event_type.name,
                    }
                    for slug, event_type in EVENT_TYPES.items()
                    if event_type.can_be_triggered_on_page_type(page_type)
                ]
                for page_type in PAGE_MODEL_CLASSES
            }
        }, cls=DjangoJSONEncoder)
    })


class StartAbTestMenuItem(ActionMenuItem):
    name = 'action-start-ab-test'
    label = _("Start A/B test")

    def is_shown(self, request, context):
        if not context['user_page_permissions'].for_page(context['ab_test'].page).can_publish():
            return False

        return context['ab_test'].status == AbTest.STATUS_DRAFT


class RestartAbTestMenuItem(ActionMenuItem):
    name = 'action-restart-ab-test'
    label = _("Restart A/B test")

    def is_shown(self, request, context):
        if not context['user_page_permissions'].for_page(context['ab_test'].page).can_publish():
            return False

        return context['ab_test'].status == AbTest.STATUS_PAUSED


class EndAbTestMenuItem(ActionMenuItem):
    name = 'action-end-ab-test'
    label = _("End A/B test")

    def is_shown(self, request, context):
        if not context['user_page_permissions'].for_page(context['ab_test'].page).can_publish():
            return False

        return context['ab_test'].status in [AbTest.STATUS_DRAFT, AbTest.STATUS_RUNNING, AbTest.STATUS_PAUSED]


class PauseAbTestMenuItem(ActionMenuItem):
    name = 'action-pause-ab-test'
    label = _("Pause A/B test")

    def is_shown(self, request, context):
        if not context['user_page_permissions'].for_page(context['ab_test'].page).can_publish():
            return False

        return context['ab_test'].status == AbTest.STATUS_RUNNING


class AbTestActionMenu:
    template = 'wagtailadmin/pages/action_menu/menu.html'

    def __init__(self, request, **kwargs):
        self.request = request
        self.context = kwargs
        self.context['user_page_permissions'] = UserPagePermissionsProxy(self.request.user)

        self.menu_items = [
            StartAbTestMenuItem(order=0),
            RestartAbTestMenuItem(order=1),
            EndAbTestMenuItem(order=2),
            PauseAbTestMenuItem(order=3),
        ]

        self.menu_items = [
            menu_item
            for menu_item in self.menu_items
            if menu_item.is_shown(self.request, self.context)
        ]

        try:
            self.default_item = self.menu_items.pop(0)
        except IndexError:
            self.default_item = None

    def render_html(self):
        return render_to_string(self.template, {
            'default_menu_item': self.default_item.render_html(self.request, self.context),
            'show_menu': bool(self.menu_items),
            'rendered_menu_items': [
                menu_item.render_html(self.request, self.context)
                for menu_item in self.menu_items
            ],
        }, request=self.request)

    @cached_property
    def media(self):
        media = forms.Media()
        for item in self.menu_items:
            media += item.media
        return media


def get_progress_and_results_common_context(request, page, ab_test):
    # Fetch stats from database
    stats = ab_test.hourly_logs.aggregate(
        control_participants=Sum('participants', filter=Q(version=AbTest.VERSION_CONTROL)),
        control_conversions=Sum('conversions', filter=Q(version=AbTest.VERSION_CONTROL)),
        variant_participants=Sum('participants', filter=Q(version=AbTest.VERSION_VARIANT)),
        variant_conversions=Sum('conversions', filter=Q(version=AbTest.VERSION_VARIANT)),
    )
    control_participants = stats['control_participants'] or 0
    control_conversions = stats['control_conversions'] or 0
    variant_participants = stats['variant_participants'] or 0
    variant_conversions = stats['variant_conversions'] or 0

    current_sample_size = control_participants + variant_participants

    estimated_completion_date = None
    if ab_test.status == AbTest.STATUS_RUNNING and current_sample_size:
        running_duration_days = ab_test.total_running_duration().days

        if running_duration_days > 0:
            participants_per_day = current_sample_size / ab_test.total_running_duration().days
            estimated_days_remaining = (ab_test.sample_size - current_sample_size) / participants_per_day
            estimated_completion_date = timezone.now().date() + datetime.timedelta(days=estimated_days_remaining)

    # Generate time series data for the chart
    time_series = []
    control = 0
    variant = 0
    date = None
    for log in ab_test.hourly_logs.order_by('date', 'hour'):
        # Accumulate the conversions
        if log.version == AbTest.VERSION_CONTROL:
            control += log.conversions
        else:
            variant += log.conversions

        while date is None or date < log.date:
            if date is None:
                # First record
                date = log.date
            else:
                # Move time forward to match log record
                date += datetime.timedelta(days=1)

            # Generate a log for this time
            time_series.append({
                'date': date,
                'control': control,
                'variant': variant,
            })

    return {
        'page': page,
        'ab_test': ab_test,
        'current_sample_size': current_sample_size,
        'current_sample_size_percent': int(current_sample_size / ab_test.sample_size * 100),
        'control_conversions': control_conversions,
        'control_participants': control_participants,
        'control_conversions_percent': int(control_conversions / control_participants * 100) if control_participants else 0,
        'variant_conversions': variant_conversions,
        'variant_participants': variant_participants,
        'variant_conversions_percent': int(variant_conversions / variant_participants * 100) if variant_participants else 0,
        'control_is_winner': ab_test.winning_version == AbTest.VERSION_CONTROL,
        'variant_is_winner': ab_test.winning_version == AbTest.VERSION_VARIANT,
        'unclear_winner': ab_test.status in [AbTest.STATUS_FINISHED, ab_test.STATUS_COMPLETED] and ab_test.winning_version is None,
        'estimated_completion_date': estimated_completion_date,
        'chart_data': json.dumps({
            'x': 'x',
            'columns': [
                ['x'] + [data_point['date'].isoformat() for data_point in time_series],
                [_("Control")] + [data_point['control'] for data_point in time_series],
                [_("Variant")] + [data_point['variant'] for data_point in time_series],
            ],
            'type': 'spline',
        }),
    }


def progress(request, page, ab_test):
    if request.method == 'POST':
        page_perms = page.permissions_for_user(request.user)

        if 'action-start-ab-test' in request.POST or 'action-restart-ab-test' in request.POST:
            if page_perms.can_publish():
                if ab_test.status in [AbTest.STATUS_DRAFT, AbTest.STATUS_PAUSED]:
                    ab_test.start()

                    messages.success(request, _("The A/B test has been started."))
                else:
                    messages.error(request, _("The A/B test must be in draft or paused in order to be started."))
            else:
                messages.error(request, _("You must have permission to publish in order to start an A/B test."))

        elif 'action-end-ab-test' in request.POST:
            if page_perms.can_publish():
                if ab_test.status in [AbTest.STATUS_DRAFT, AbTest.STATUS_RUNNING, AbTest.STATUS_PAUSED]:
                    ab_test.cancel()

                    messages.success(request, _("The A/B test has been ended."))
                elif ab_test.status == AbTest.STATUS_FINISHED:
                    ab_test.complete(AbTest.COMPLETION_ACTION_DO_NOTHING, user=request.user)
                else:
                    messages.error(request, _("The A/B test has already ended."))
            else:
                messages.error(request, _("You must have permission to publish in order to end an A/B test."))

        elif 'action-pause-ab-test' in request.POST:
            if page_perms.can_publish():
                if ab_test.status == AbTest.STATUS_RUNNING:
                    ab_test.pause()

                    messages.success(request, _("The A/B test has been paused."))
                else:
                    messages.error(request, _("The A/B test cannot be paused because it is not running."))
            else:
                messages.error(request, _("You must have permission to publish in order to pause an A/B test."))

        elif 'action-select-control' in request.POST:
            if ab_test.status == AbTest.STATUS_FINISHED:
                ab_test.complete(AbTest.COMPLETION_ACTION_REVERT, user=request.user)

                messages.success(request, _("The page has been reverted back to the control version."))

            else:
                messages.error(request, _("The A/B test cannot be paused because it is not running."))

        elif 'action-select-variant' in request.POST:
            if ab_test.status == AbTest.STATUS_FINISHED:
                # TODO Permission check?
                ab_test.complete(AbTest.COMPLETION_ACTION_PUBLISH, user=request.user)

                messages.success(request, _("The variant version has been published."))

            else:
                messages.error(request, _("The A/B test cannot be paused because it is not running."))

        else:
            messages.error(request, _("Unknown action"))

        # Redirect back
        return redirect('wagtailadmin_pages:edit', page.id)

    context = get_progress_and_results_common_context(request, page, ab_test)
    context['action_menu'] = AbTestActionMenu(request, view='edit', page=page, ab_test=ab_test)
    return render(request, 'wagtail_ab_testing/progress.html', context)


def results(request, page_id, ab_test_id):
    page = get_object_or_404(Page, id=page_id)
    if not page.permissions_for_user(request.user).can_edit():
        raise PermissionDenied

    ab_test = get_object_or_404(AbTest, page=page, id=ab_test_id)

    context = get_progress_and_results_common_context(request, page, ab_test)
    return render(request, 'wagtail_ab_testing/results.html', context)


def compare_draft(request, page_id):
    page = get_object_or_404(Page, id=page_id).specific

    latest_revision_as_page = page.get_latest_revision().as_page_object()
    comparison = page.get_edit_handler().get_comparison()
    comparison = [comp(page, latest_revision_as_page) for comp in comparison]
    comparison = [comp for comp in comparison if comp.has_changed()]

    return render(request, 'wagtail_ab_testing/compare.html', {
        'page': page,
        'latest_revision_as_page': latest_revision_as_page,
        'comparison': comparison,
    })


# TEMPORARY
def add_test_participants(request, ab_test_id):
    ab_test = get_object_or_404(AbTest, id=ab_test_id)

    for i in range(int(ab_test.sample_size / 10)):
        ab_test.add_participant()

    return redirect('wagtailadmin_pages:edit', ab_test.page_id)


def add_test_conversions(request, ab_test_id, version):
    ab_test = get_object_or_404(AbTest, id=ab_test_id)

    for i in range(int(ab_test.sample_size / 10)):
        ab_test.log_conversion(version, time=timezone.now() - datetime.timedelta(days=random.randint(1, 20), hours=random.randint(0, 24)))

    return redirect('wagtailadmin_pages:edit', ab_test.page_id)


class SearchPageTitleFilter(django_filters.CharFilter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        return qs.filter(page__title__icontains=value)


class AbTestingReportFilterSet(WagtailFilterSet):
    page = SearchPageTitleFilter()
    first_started_at = django_filters.DateFromToRangeFilter(label=gettext_lazy("Started at"), widget=DateRangePickerWidget)

    class Meta:
        model = AbTest
        fields = ['status', 'page', 'first_started_at']


class AbTestingReportView(ReportView):
    template_name = 'wagtail_ab_testing/report.html'
    title = gettext_lazy('A/B testing')
    header_icon = ''

    filterset_class = AbTestingReportFilterSet

    def get_queryset(self):
        return AbTest.objects.all().order_by(F('first_started_at').desc(nulls_first=True))



import datetime

from django.test import TestCase
from freezegun import freeze_time
from wagtail.core.models import Page

from wagtail_ab_testing.models import AbTest, AbTestHourlyLog


@freeze_time('2020-11-04T22:37:00Z')
class TestAbTestModel(TestCase):
    def setUp(self):
        home_page = Page.objects.get(id=2)
        home_page.title = "Changed title"
        revision = home_page.save_revision()
        self.ab_test = AbTest.objects.create(
            page=home_page,
            name="Test",
            variant_revision=revision,
            goal_event="foo",
            sample_size=10,
        )

    def test_finish(self):
        self.ab_test.finish()
        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_FINISHED)

    def test_cancel(self):
        self.ab_test.cancel()
        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)

    def test_add_participant(self):
        version = self.ab_test.add_participant()

        # This should've created a history log
        log = self.ab_test.hourly_logs.get()

        self.assertEqual(log.date, datetime.date(2020, 11, 4))
        self.assertEqual(log.hour, 22)
        self.assertEqual(log.version, version)
        self.assertEqual(log.participants, 1)
        self.assertEqual(log.conversions, 0)

    def test_log_conversion(self):
        self.ab_test.log_conversion(AbTest.VERSION_CONTROL)

        # This should've created a history log
        log = self.ab_test.hourly_logs.get()

        self.assertEqual(log.date, datetime.date(2020, 11, 4))
        self.assertEqual(log.hour, 22)
        self.assertEqual(log.version, AbTest.VERSION_CONTROL)
        self.assertEqual(log.participants, 0)
        self.assertEqual(log.conversions, 1)

        # Now add another
        self.ab_test.log_conversion(AbTest.VERSION_CONTROL)

        log.refresh_from_db()

        self.assertEqual(log.participants, 0)
        self.assertEqual(log.conversions, 2)

    def set_up_test(self, control_participants, control_conversions, variant_participants, variant_conversions):
        AbTestHourlyLog.objects.create(
            ab_test=self.ab_test,
            version=AbTest.VERSION_CONTROL,
            date=datetime.date(2020, 11, 4),
            hour=22,
            participants=control_participants,
            conversions=control_conversions,
        )

        AbTestHourlyLog.objects.create(
            ab_test=self.ab_test,
            version=AbTest.VERSION_VARIANT,
            date=datetime.date(2020, 11, 4),
            hour=22,
            participants=variant_participants,
            conversions=variant_conversions,
        )

    def test_check_for_winner_no_data(self):
        self.set_up_test(0, 0, 0, 0)

        self.assertIsNone(self.ab_test.check_for_winner())

    def test_check_control_clearly_wins(self):
        self.set_up_test(100, 80, 100, 20)

        self.assertEqual(self.ab_test.check_for_winner(), AbTest.VERSION_CONTROL)

    def test_check_variantarly_wins(self):
        self.set_up_test(100, 20, 100, 80)

        self.assertEqual(self.ab_test.check_for_winner(), AbTest.VERSION_VARIANT)

    def test_control_just_wins(self):
        self.set_up_test(100, 64, 100, 50)

        self.assertEqual(self.ab_test.check_for_winner(), AbTest.VERSION_CONTROL)

    def test_variantt_wins(self):
        self.set_up_test(100, 50, 100, 64)

        self.assertEqual(self.ab_test.check_for_winner(), AbTest.VERSION_VARIANT)

    def test_close_leaning_control(self):
        self.set_up_test(100, 62, 100, 50)

        self.assertIsNone(self.ab_test.check_for_winner())

    def test_close_leaning_variant(self):
        self.set_up_test(100, 50, 100, 62)

        self.assertIsNone(self.ab_test.check_for_winner())

    def test_confidence_improves_with_more_participants(self):
        # Even though as a percentage, this is less of a win than in previous tests,
        # we can be more confident with a slight difference if there are more paricipants
        self.set_up_test(1000, 550, 1000, 500)

        self.assertEqual(self.ab_test.check_for_winner(), AbTest.VERSION_CONTROL)


class TestAutoCancelOnUnpublish(TestCase):
    def setUp(self):
        self.home_page = Page.objects.get(id=2)
        self.home_page.title = "Changed title"
        revision = self.home_page.save_revision()
        self.ab_test = AbTest.objects.create(
            page=self.home_page,
            name="Test",
            variant_revision=revision,
            goal_event="foo",
            sample_size=10,
        )

    def test_unpublish_draft(self):
        self.ab_test.status = AbTest.STATUS_DRAFT
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)

    def test_unpublish_running(self):
        self.ab_test.status = AbTest.STATUS_RUNNING
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)

    def test_unpublish_paused(self):
        self.ab_test.status = AbTest.STATUS_PAUSED
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)

    def test_unpublish_finished(self):
        self.ab_test.status = AbTest.STATUS_FINISHED
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_COMPLETED)

    def test_unpublish_completed(self):
        self.ab_test.status = AbTest.STATUS_COMPLETED
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_COMPLETED)

    def test_unpublish_cancelled(self):
        self.ab_test.status = AbTest.STATUS_CANCELLED
        self.ab_test.save()

        self.home_page.unpublish()

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)



from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.urls import reverse
from wagtail.core.models import Page
from wagtail.tests.utils import WagtailTestUtils

from wagtail_ab_testing.models import AbTest
from wagtail_ab_testing.test.models import SimplePage


class TestProgressView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Convert the user into an moderator
        self.moderators_group = Group.objects.get(name="Moderators")
        for permission in Permission.objects.filter(content_type=ContentType.objects.get_for_model(AbTest)):
            self.moderators_group.permissions.add(permission)
        self.user.is_superuser = False
        self.user.groups.add(self.moderators_group)
        self.user.save()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()

        # Edit the page and save a revision
        self.page.title = "Test updated"
        revision = self.page.save_revision()

        # Create an A/B test
        self.ab_test = AbTest.objects.create(
            page=self.page,
            name="Test",
            variant_revision=revision,
            status=AbTest.STATUS_RUNNING,
            sample_size=100,
        )

    def test_get_progress(self):
        response = self.client.get(reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.assertNotContains(response, "Save draft")
        self.assertTemplateUsed(response, "wagtail_ab_testing/progress.html")

    def test_post_start(self):
        self.ab_test.status = AbTest.STATUS_DRAFT
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-start-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_RUNNING)

    def test_post_pause(self):
        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-pause-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_PAUSED)

    def test_post_restart(self):
        self.ab_test.status = AbTest.STATUS_PAUSED
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-restart-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_RUNNING)

    def test_post_end(self):
        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-end-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_CANCELLED)

    def test_post_start_without_publish_permission(self):
        self.moderators_group.page_permissions.filter(permission_type='publish').delete()

        self.ab_test.status = AbTest.STATUS_DRAFT
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-start-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_DRAFT)

    def test_post_pause_without_publish_permission(self):
        self.moderators_group.page_permissions.filter(permission_type='publish').delete()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-pause-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_RUNNING)

    def test_post_restart_without_publish_permission(self):
        self.moderators_group.page_permissions.filter(permission_type='publish').delete()

        self.ab_test.status = AbTest.STATUS_PAUSED
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-restart-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_PAUSED)

    def test_post_end_without_publish_permission(self):
        self.moderators_group.page_permissions.filter(permission_type='publish').delete()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-end-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_RUNNING)

    def test_post_end_when_finished(self):
        self.ab_test.status = AbTest.STATUS_FINISHED
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-end-ab-test': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()

        self.assertEqual(self.ab_test.status, AbTest.STATUS_COMPLETED)
        self.assertEqual(self.ab_test.page.title, "Test")
        self.assertTrue(self.ab_test.page.has_unpublished_changes)

    def test_post_select_control(self):
        self.ab_test.status = AbTest.STATUS_FINISHED
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-select-control': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_COMPLETED)
        self.assertEqual(self.ab_test.page.title, "Test")
        self.assertFalse(self.ab_test.page.has_unpublished_changes)

    def test_post_select_variant(self):
        self.ab_test.status = AbTest.STATUS_FINISHED
        self.ab_test.save()

        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'action-select-variant': 'on',
        })

        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        self.ab_test.refresh_from_db()
        self.assertEqual(self.ab_test.status, AbTest.STATUS_COMPLETED)
        self.assertEqual(self.ab_test.page.title, "Test updated")
        self.assertFalse(self.ab_test.page.has_unpublished_changes)



import json

from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.urls import reverse
from wagtail.core.models import Page
from wagtail.tests.utils import WagtailTestUtils

from wagtail_ab_testing.models import AbTest
from wagtail_ab_testing.test.models import SimplePage

from .utils import assert_permission_denied


class TestSaveAndCreateAbTestButton(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Convert the user into an moderator
        self.moderators_group = Group.objects.get(name="Moderators")
        for permission in Permission.objects.filter(content_type=ContentType.objects.get_for_model(AbTest)):
            self.moderators_group.permissions.add(permission)
        self.user.is_superuser = False
        self.user.groups.add(self.moderators_group)
        self.user.save()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()

    def test_shows_on_page_edit(self):
        response = self.client.get(reverse('wagtailadmin_pages:edit', args=[self.page.id]))
        self.assertContains(response, "Save and create A/B Test")

    def test_click_on_page_edit(self):
        response = self.client.post(reverse('wagtailadmin_pages:edit', args=[self.page.id]), {
            'title': "Test version",
            'slug': "test",
            'create-ab-test': '',
        })
        self.assertRedirects(response, reverse('wagtail_ab_testing:add_ab_test_compare', args=[self.page.id]))

    def test_doesnt_show_on_page_create(self):
        response = self.client.get(reverse('wagtailadmin_pages:add', args=["wagtail_ab_testing_test", "simplepage", self.page.id]))
        self.assertNotContains(response, "Save and create A/B Test")


# These tests are required for both the compare and form views
class PermissionTests:
    def test_without_page_edit_permission(self):
        self.moderators_group.page_permissions.all().delete()

        response = self.get(self.page.id)
        assert_permission_denied(self, response)

    def test_without_add_abtest_permission(self):
        add_abtest_permission = Permission.objects.get(codename='add_abtest')
        self.moderators_group.permissions.remove(add_abtest_permission)

        response = self.get(self.page.id)
        assert_permission_denied(self, response)

    def _create_abtest(self, status):
        AbTest.objects.create(
            page=self.page,
            name="Test",
            variant_revision=self.page.get_latest_revision(),
            status=status,
            sample_size=100,
        )

    def test_with_existing_draft_abtest(self):
        self._create_abtest(AbTest.STATUS_DRAFT)

        response = self.get(self.page.id)
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

    def test_with_existing_running_abtest(self):
        self._create_abtest(AbTest.STATUS_RUNNING)

        response = self.get(self.page.id)
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

    def test_with_existing_paused_abtest(self):
        self._create_abtest(AbTest.STATUS_PAUSED)

        response = self.get(self.page.id)
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

    def test_with_existing_cancelled_abtest(self):
        self._create_abtest(AbTest.STATUS_CANCELLED)

        response = self.get(self.page.id)
        self.assertEqual(response.status_code, 200)

    def test_with_existing_finished_abtest(self):
        self._create_abtest(AbTest.STATUS_FINISHED)

        response = self.get(self.page.id)
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

    def test_with_existing_completed_abtest(self):
        self._create_abtest(AbTest.STATUS_COMPLETED)

        response = self.get(self.page.id)
        self.assertEqual(response.status_code, 200)


class TestAddAbTestCompareView(WagtailTestUtils, TestCase, PermissionTests):
    def setUp(self):
        self.user = self.login()

        # Convert the user into an moderator
        self.moderators_group = Group.objects.get(name="Moderators")
        for permission in Permission.objects.filter(content_type=ContentType.objects.get_for_model(AbTest)):
            self.moderators_group.permissions.add(permission)
        self.user.is_superuser = False
        self.user.groups.add(self.moderators_group)
        self.user.save()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()
        self.page.title = "Test version"
        self.latest_revision = self.page.save_revision()

    def get(self, page_id):
        return self.client.get(reverse('wagtail_ab_testing:add_ab_test_compare', args=[page_id]))

    def test_get_add_compare(self):
        response = self.get(self.page.id)
        self.assertEqual(response.status_code, 200)


class TestAddAbTestFormView(WagtailTestUtils, TestCase, PermissionTests):
    def setUp(self):
        self.user = self.login()

        # Convert the user into an moderator
        self.moderators_group = Group.objects.get(name="Moderators")
        for permission in Permission.objects.filter(content_type=ContentType.objects.get_for_model(AbTest)):
            self.moderators_group.permissions.add(permission)
        self.user.is_superuser = False
        self.user.groups.add(self.moderators_group)
        self.user.save()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()
        self.page.title = "Test version"
        self.latest_revision = self.page.save_revision()

    def get(self, page_id):
        return self.client.get(reverse('wagtail_ab_testing:add_ab_test_form', args=[page_id]))

    def test_get_add_form(self):
        response = self.get(self.page.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.context['goal_selector_props']), {
            "goalTypesByPageType": {
                "wagtailcore.page": [{"slug": "visit-page", "name": "Visit page"}],
                "wagtail_ab_testing_test.simplepage": [{"slug": "visit-page", "name": "Visit page"}]
            }
        })

    def test_post_add_form(self):
        response = self.client.post(reverse('wagtail_ab_testing:add_ab_test_form', args=[self.page.id]), {
            'name': 'Test',
            'hypothesis': 'Does changing the title to "Donate now!" increase donations?',
            'goal_event': 'visit-page',
            'goal_page': '',
            'sample_size': '100'
        })
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        ab_test = AbTest.objects.get()
        self.assertEqual(ab_test.page, self.page.page_ptr)
        self.assertEqual(ab_test.variant_revision, self.latest_revision)
        self.assertEqual(ab_test.name, 'Test')
        self.assertEqual(ab_test.hypothesis, 'Does changing the title to "Donate now!" increase donations?')
        self.assertEqual(ab_test.goal_event, 'visit-page')
        self.assertIsNone(ab_test.goal_page)
        self.assertEqual(ab_test.sample_size, 100)
        self.assertEqual(ab_test.created_by, self.user)
        self.assertEqual(ab_test.status, AbTest.STATUS_DRAFT)

    def test_post_add_form_start(self):
        response = self.client.post(reverse('wagtail_ab_testing:add_ab_test_form', args=[self.page.id]), {
            'name': 'Test',
            'hypothesis': 'Does changing the title to "Donate now!" increase donations?',
            'goal_event': 'visit-page',
            'goal_page': '',
            'sample_size': '100',
            'start': 'on'
        })
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        ab_test = AbTest.objects.get()
        self.assertEqual(ab_test.status, AbTest.STATUS_RUNNING)

    def test_post_add_form_start_without_publish_permission(self):
        self.moderators_group.page_permissions.filter(permission_type='publish').delete()

        response = self.client.post(reverse('wagtail_ab_testing:add_ab_test_form', args=[self.page.id]), {
            'name': 'Test',
            'hypothesis': 'Does changing the title to "Donate now!" increase donations?',
            'goal_event': 'visit-page',
            'goal_page': '',
            'sample_size': '100',
            'start': 'on'
        })
        self.assertRedirects(response, reverse('wagtailadmin_pages:edit', args=[self.page.id]))

        ab_test = AbTest.objects.get()
        self.assertEqual(ab_test.status, AbTest.STATUS_DRAFT)



from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.urls import reverse
from wagtail.core.models import Page
from wagtail.tests.utils import WagtailTestUtils

from wagtail_ab_testing.models import AbTest
from wagtail_ab_testing.test.models import SimplePage


class TestResultsView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Convert the user into an moderator
        self.moderators_group = Group.objects.get(name="Moderators")
        for permission in Permission.objects.filter(content_type=ContentType.objects.get_for_model(AbTest)):
            self.moderators_group.permissions.add(permission)
        self.user.is_superuser = False
        self.user.groups.add(self.moderators_group)
        self.user.save()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()

        # Create an A/B test
        self.ab_test = AbTest.objects.create(
            page=self.page,
            name="Test",
            variant_revision=self.page.get_latest_revision(),
            status=AbTest.STATUS_COMPLETED,
            sample_size=100,
        )

    def test_get_results(self):
        response = self.client.get(reverse('wagtail_ab_testing:results', args=[self.page.id, self.ab_test.id]))

        self.assertTemplateUsed(response, "wagtail_ab_testing/results.html")



from django.utils.translation import gettext_lazy as __
from wagtail.contrib.forms.utils import get_form_types


class BaseEvent:
    """
    A base class for events that are linked to Wagtail pages.
    """
    name = None

    # A list of page model classes that this event can be triggered on.
    # This may be overridden by the .get_page_types() method
    page_types = None

    def get_page_types(self):
        """
        Returns a list of page model classes that this event can be triggered on.
        """
        return self.page_types

    def can_be_triggered_on_page_type(self, page_type):
        """
        Returns True if this event can be triggered on an instance of the given page model.
        """
        page_types = self.get_page_types()

        if page_types is None:
            return True

        return page_type in page_types


class VisitPageEvent(BaseEvent):
    """
    Triggered when a user visits a page.
    """
    name = __("Visit page")


class SubmitFormPageEvent(BaseEvent):
    """
    Triggered when a user submits a form page.
    """
    name = __("Submit form page")

    def get_page_types(self):
        return [
            ct.model_class()
            for ct in get_form_types()
        ]


EVENT_TYPES = {
    'visit-page': VisitPageEvent(),
    'submit-form': SubmitFormPageEvent(),
}



from django.test import TestCase
from wagtail.core.models import Page

from wagtail_ab_testing.models import AbTest


class TestServe(TestCase):
    def setUp(self):
        self.home_page = Page.objects.get(id=2)
        self.home_page.title = "Changed title"
        revision = self.home_page.save_revision()
        self.ab_test = AbTest.objects.create(
            page=self.home_page,
            name="Test",
            variant_revision=revision,
            goal_event="foo",
            sample_size=10,
        )

    def test_serves_control(self):
        # Add a participant for variant
        # This will make the new participant use control to balance the numbers
        self.ab_test.add_participant(AbTest.VERSION_VARIANT)

        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)

        self.assertContains(response, "Welcome to your new Wagtail site!")
        self.assertNotContains(response, "Changed title")

        self.assertEqual(self.client.session[f'wagtail-ab-testing_{self.ab_test.id}_version'], AbTest.VERSION_CONTROL)

    def test_serves_variant(self):
        # Add a participant for control
        # This will make the new participant use variant to balance the numbers
        self.ab_test.add_participant(AbTest.VERSION_CONTROL)

        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)

        self.assertNotContains(response, "Welcome to your new Wagtail site!")
        self.assertContains(response, "Changed title")

        self.assertEqual(self.client.session[f'wagtail-ab-testing_{self.ab_test.id}_version'], AbTest.VERSION_VARIANT)

    def test_doesnt_track_bots(self):
        # Add a participant for control
        # This will make it serve the variant if it does incorrectly decide to track the user
        self.ab_test.add_participant(AbTest.VERSION_CONTROL)

        response = self.client.get(
            '/',
            HTTP_USER_AGENT='Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)'
        )
        self.assertEqual(response.status_code, 200)

        # The control should be served
        self.assertContains(response, "Welcome to your new Wagtail site!")
        self.assertNotContains(response, "Changed title")
        self.assertNotIn(f'wagtail-ab-testing_{self.ab_test.id}_version', self.client.session)

    def test_doesnt_track_dnt_users(self):
        # Add a participant for control
        # This will make it serve the variant if it does incorrectly decide to track the user
        self.ab_test.add_participant(AbTest.VERSION_CONTROL)

        response = self.client.get('/', HTTP_DNT='1')
        self.assertEqual(response.status_code, 200)

        # The control should be served
        self.assertContains(response, "Welcome to your new Wagtail site!")
        self.assertNotContains(response, "Changed title")
        self.assertNotIn(f'wagtail-ab-testing_{self.ab_test.id}_version', self.client.session)



from django.apps import AppConfig


class WagtailAbTestingTestAppConfig(AppConfig):
    label = "wagtail_ab_testing_test"
    name = "wagtail_ab_testing.test"
    verbose_name = "Wagtail A/B Testing tests"



from django.test import TestCase
from django.urls import reverse
from wagtail.core.models import Page
from wagtail.tests.utils import WagtailTestUtils

from wagtail_ab_testing.models import AbTest
from wagtail_ab_testing.test.models import SimplePage


class TestCompareDraftView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.user = self.login()

        # Create test page with a draft revision
        self.page = Page.objects.get(id=1).add_child(instance=SimplePage(title="Test", slug="test"))
        self.page.save_revision().publish()

        # Create an A/B test
        self.ab_test = AbTest.objects.create(
            page=self.page,
            name="Test",
            variant_revision=self.page.get_latest_revision(),
            status=AbTest.STATUS_RUNNING,
            sample_size=100,
        )

    def test_get_compare_draft(self):
        response = self.client.get(reverse('wagtail_ab_testing:compare_draft', args=[self.page.id]))

        self.assertTemplateUsed(response, "wagtail_ab_testing/compare.html")



import json

from django.contrib.auth.models import Permission
from django.shortcuts import redirect
from django.urls import path, include, reverse
from django.utils.html import format_html, escapejs
from django.utils.translation import gettext as _, gettext_lazy as __
from django.views.i18n import JavaScriptCatalog

from wagtail.admin.action_menu import ActionMenuItem
from wagtail.admin.menu import MenuItem
from wagtail.admin.staticfiles import versioned_static
from wagtail.core import hooks

from . import views
from .compat import DATE_FORMAT
from .models import AbTest
from .utils import request_is_trackable


@hooks.register("register_admin_urls")
def register_admin_urls():
    urls = [
        path('jsi18n/', JavaScriptCatalog.as_view(packages=['wagtail_ab_testing']), name='javascript_catalog'),
        path('add/<int:page_id>/compare/', views.add_compare, name='add_ab_test_compare'),
        path('<int:page_id>/compare-draft/', views.compare_draft, name='compare_draft'),
        path('add/<int:page_id>/', views.add_form, name='add_ab_test_form'),
        path('add-test-participants/<int:ab_test_id>/', views.add_test_participants, name='add_test_participants'),
        path('add-test-conversions/<int:ab_test_id>/<slug:version>', views.add_test_conversions, name='add_test_conversions'),
        path('report/', views.AbTestingReportView.as_view(), name='report'),
        path('results/<int:page_id>/<int:ab_test_id>/', views.results, name='results'),
    ]

    return [
        path(
            "abtests/",
            include(
                (urls, "wagtail_ab_testing"),
                namespace="wagtail_ab_testing",
            ),
        )
    ]


class CreateAbTestActionMenuItem(ActionMenuItem):
    name = 'create-ab-test'
    label = __("Save and create A/B Test")
    icon_name = 'people-arrows'

    def is_shown(self, request, context):
        if context['view'] != 'edit':
            return False

        # User must have permission to add A/B tests
        if not request.user.has_perm('wagtail_ab_testing.add_abtest'):
            return False

        return True


@hooks.register('register_page_action_menu_item')
def register_create_abtest_action_menu_item():
    return CreateAbTestActionMenuItem(order=100)


# This is the only way to inject custom JS into the editor with knowledge of the page being edited
class AbTestingTabActionMenuItem(ActionMenuItem):
    def render_html(self, request, context):
        if 'page' in context:
            return format_html(
                '<script src="{}"></script><script src="{}"></script><script>window.abTestingTabProps = JSON.parse("{}");</script>',
                reverse('wagtail_ab_testing:javascript_catalog'),
                versioned_static('wagtail_ab_testing/js/wagtail-ab-testing.js'),
                escapejs(json.dumps({
                    'tests': [
                        {
                            'id': ab_test.id,
                            'name': ab_test.name,
                            'started_at': ab_test.first_started_at.strftime(DATE_FORMAT) if ab_test.first_started_at else _("Not started"),
                            'status': ab_test.get_status_description(),
                            'results_url': reverse('wagtail_ab_testing:results', args=[ab_test.page_id, ab_test.id]),
                        }
                        for ab_test in AbTest.objects.filter(page=context['page']).order_by('-id')
                    ],
                    'can_create_abtest': request.user.has_perm('wagtail_ab_testing.add_abtest'),
                }))
            )

        return ''


@hooks.register('register_page_action_menu_item')
def register_ab_testing_tab_action_menu_item():
    return AbTestingTabActionMenuItem()


@hooks.register('after_edit_page')
def redirect_to_create_ab_test(request, page):
    if 'create-ab-test' in request.POST:
        return redirect('wagtail_ab_testing:add_ab_test_compare', page.id)


@hooks.register('before_edit_page')
def check_for_running_ab_test(request, page):
    running_experiment = AbTest.objects.get_current_for_page(page=page)
    if running_experiment:
        return views.progress(request, page, running_experiment)


@hooks.register('before_serve_page')
def before_serve_page(page, request, serve_args, serve_kwargs):
    # Check if there are any running tests on the page
    try:
        test = AbTest.objects.get(page=page, status=AbTest.STATUS_DRAFT)
    except AbTest.DoesNotExist:
        return

    # Check if the user is trackable
    if not request_is_trackable(request):
        return

    # Make the user a participant if they're not already
    if f'wagtail-ab-testing_{test.id}_version' not in request.session:
        request.session[f'wagtail-ab-testing_{test.id}_version'] = test.add_participant()

    # If the user is visiting the variant version, serve that from the revision
    if request.session[f'wagtail-ab-testing_{test.id}_version'] == AbTest.VERSION_VARIANT:
        return test.variant_revision.as_page_object().serve(request, *serve_args, **serve_kwargs)


class AbTestingReportMenuItem(MenuItem):

    def is_shown(self, request):
        return True


@hooks.register('register_reports_menu_item')
def register_ab_testing_report_menu_item():
    return AbTestingReportMenuItem(_('A/B testing'), reverse('wagtail_ab_testing:report'), icon_name='people-arrows', order=1000)


@hooks.register('register_icons')
def register_icons(icons):
    icons.append('wagtail_ab_testing/icons/people-arrows.svg')
    icons.append('wagtail_ab_testing/icons/crown.svg')
    return icons


@hooks.register('register_permissions')
def register_add_abtest_permission():
    return Permission.objects.filter(content_type__app_label='wagtail_ab_testing', codename='add_abtest')
